import commons_pb2 as _commons_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class InferRequest(_message.Message):
    __slots__ = ('argument',)
    ARGUMENT_FIELD_NUMBER: _ClassVar[int]
    argument: _commons_pb2.Variant

    def __init__(self, argument: _Optional[_Union[_commons_pb2.Variant, _Mapping]]=...) -> None:
        ...

class InferResponse(_message.Message):
    __slots__ = ('prediction',)
    PREDICTION_FIELD_NUMBER: _ClassVar[int]
    prediction: _commons_pb2.Variant

    def __init__(self, prediction: _Optional[_Union[_commons_pb2.Variant, _Mapping]]=...) -> None:
        ...


from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class VariantType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NONE: _ClassVar[VariantType]
    DOUBLE: _ClassVar[VariantType]
    INT: _ClassVar[VariantType]
    STRING: _ClassVar[VariantType]
    PARQUET: _ClassVar[VariantType]
    ARROW: _ClassVar[VariantType]
    JSON: _ClassVar[VariantType]
NONE: VariantType
DOUBLE: VariantType
INT: VariantType
STRING: VariantType
PARQUET: VariantType
ARROW: VariantType
JSON: VariantType

class Variant(_message.Message):
    __slots__ = ('type', 'value')
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    type: VariantType
    value: bytes

    def __init__(self, type: _Optional[_Union[VariantType, str]]=..., value: _Optional[bytes]=...) -> None:
        ...

class Argument(_message.Message):
    __slots__ = ('position', 'data')
    POSITION_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    position: int
    data: Variant

    def __init__(self, position: _Optional[int]=..., data: _Optional[_Union[Variant, _Mapping]]=...) -> None:
        ...

class KwArgument(_message.Message):
    __slots__ = ('keyword', 'data')
    KEYWORD_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    keyword: str
    data: Variant

    def __init__(self, keyword: _Optional[str]=..., data: _Optional[_Union[Variant, _Mapping]]=...) -> None:
        ...

class Status(_message.Message):
    __slots__ = ('code', 'message')
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: str
    message: str

    def __init__(self, code: _Optional[str]=..., message: _Optional[str]=...) -> None:
        ...


import json
import struct
import io
import pandas
import pyarrow.parquet as pq
import pyarrow as pa
from model_runner.grpc.generated.commons_pb2 import VariantType


# Encoder: Converts data to bytes
def encode_data(data_type: VariantType, data) -> bytes:
    if data_type == VariantType.DOUBLE:
        return struct.pack("d", data)
    elif data_type == VariantType.INT:
        return data.to_bytes(8, byteorder="little", signed=True)
    elif data_type == VariantType.STRING:
        return data.encode("utf-8")
    elif data_type == VariantType.PARQUET:
        table = pa.Table.from_pandas(data) if isinstance(data, pandas.DataFrame) else data
        sink = io.BytesIO()
        pq.write_table(table, sink)
        return sink.getvalue()
    elif data_type == VariantType.JSON:
        try:
            json_data = json.dumps(data)  # Convert the object to a JSON string
            return json_data.encode("utf-8")  # Return the JSON string as bytes
        except TypeError as e:
            raise ValueError(f"Data cannot be serialized to JSON: {e}")
    else:
        raise ValueError(f"Unsupported data type: {data_type}")


# Decoder: Converts bytes to data
def decode_data(data_bytes: bytes, data_type: VariantType):
    if data_type == VariantType.DOUBLE:
        return struct.unpack("d", data_bytes)[0]
    elif data_type == VariantType.INT:
        return int.from_bytes(data_bytes, byteorder="little", signed=True)
    elif data_type == VariantType.STRING:
        return data_bytes.decode("utf-8")
    elif data_type == VariantType.PARQUET:
        buffer = io.BytesIO(data_bytes)
        return pq.read_table(buffer).to_pandas()
    elif data_type == VariantType.JSON:
        try:
            json_data = data_bytes.decode("utf-8")
            return json.loads(json_data)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to decode JSON data: {e}")

    else:
        raise ValueError(f"Unsupported data type: {data_type}")


def detect_data_type(data) -> VariantType:
    """
    Detects the data type based on the Python object and returns
    the corresponding VariantType enum.
    """
    if isinstance(data, float):
        return VariantType.DOUBLE
    elif isinstance(data, int):
        return VariantType.INT
    elif isinstance(data, str):
        return VariantType.STRING
    elif isinstance(data, pandas.DataFrame):
        return VariantType.PARQUET
    elif isinstance(data, dict) or isinstance(data, list):
        return VariantType.JSON
    else:
        raise ValueError(f"Unsupported data type: {type(data)}")



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 29, 0, '', 'train_infer.proto')
_sym_db = _symbol_database.Default()
from . import commons_pb2 as commons__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11train_infer.proto\x12\x0btrain_infer\x1a\rcommons.proto\x1a\x1bgoogle/protobuf/empty.proto"2\n\x0cInferRequest\x12"\n\x08argument\x18\x01 \x01(\x0b2\x10.commons.Variant"5\n\rInferResponse\x12$\n\nprediction\x18\x01 \x01(\x0b2\x10.commons.Variant2\xd2\x01\n\x17TrainInferStreamService\x127\n\x05Setup\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\x12>\n\x05Infer\x12\x19.train_infer.InferRequest\x1a\x1a.train_infer.InferResponse\x12>\n\x0cReinitialize\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty2\xcc\x01\n\x11TrainInferService\x127\n\x05Setup\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\x12>\n\x05Infer\x12\x19.train_infer.InferRequest\x1a\x1a.train_infer.InferResponse\x12>\n\x0cReinitialize\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Emptyb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'train_infer_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_INFERREQUEST']._serialized_start = 78
    _globals['_INFERREQUEST']._serialized_end = 128
    _globals['_INFERRESPONSE']._serialized_start = 130
    _globals['_INFERRESPONSE']._serialized_end = 183
    _globals['_TRAININFERSTREAMSERVICE']._serialized_start = 186
    _globals['_TRAININFERSTREAMSERVICE']._serialized_end = 396
    _globals['_TRAININFERSERVICE']._serialized_start = 399
    _globals['_TRAININFERSERVICE']._serialized_end = 603


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 29, 0, '', 'dynamic_subclass.proto')
_sym_db = _symbol_database.Default()
from . import commons_pb2 as commons__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16dynamic_subclass.proto\x12\x10dynamic_subclass\x1a\rcommons.proto\x1a\x1bgoogle/protobuf/empty.proto"\x81\x01\n\x0cSetupRequest\x12\x11\n\tclassName\x18\x01 \x01(\t\x12,\n\x11instanceArguments\x18\x02 \x03(\x0b2\x11.commons.Argument\x120\n\x13instanceKwArguments\x18\x03 \x03(\x0b2\x13.commons.KwArgument"0\n\rSetupResponse\x12\x1f\n\x06status\x18\x01 \x01(\x0b2\x0f.commons.Status"}\n\x0bCallRequest\x12\x12\n\nmethodName\x18\x01 \x01(\t\x12*\n\x0fmethodArguments\x18\x02 \x03(\x0b2\x11.commons.Argument\x12.\n\x11methodKwArguments\x18\x03 \x03(\x0b2\x13.commons.KwArgument"Y\n\x0cCallResponse\x12\x1f\n\x06status\x18\x01 \x01(\x0b2\x0f.commons.Status\x12(\n\x0emethodResponse\x18\x02 \x01(\x0b2\x10.commons.Variant"/\n\x0cRestResponse\x12\x1f\n\x06status\x18\x01 \x01(\x0b2\x0f.commons.Status2\xe9\x01\n\x16DynamicSubclassService\x12H\n\x05Setup\x12\x1e.dynamic_subclass.SetupRequest\x1a\x1f.dynamic_subclass.SetupResponse\x12E\n\x04Call\x12\x1d.dynamic_subclass.CallRequest\x1a\x1e.dynamic_subclass.CallResponse\x12>\n\x04Rest\x12\x16.google.protobuf.Empty\x1a\x1e.dynamic_subclass.RestResponseb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dynamic_subclass_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_SETUPREQUEST']._serialized_start = 89
    _globals['_SETUPREQUEST']._serialized_end = 218
    _globals['_SETUPRESPONSE']._serialized_start = 220
    _globals['_SETUPRESPONSE']._serialized_end = 268
    _globals['_CALLREQUEST']._serialized_start = 270
    _globals['_CALLREQUEST']._serialized_end = 395
    _globals['_CALLRESPONSE']._serialized_start = 397
    _globals['_CALLRESPONSE']._serialized_end = 486
    _globals['_RESTRESPONSE']._serialized_start = 488
    _globals['_RESTRESPONSE']._serialized_end = 535
    _globals['_DYNAMICSUBCLASSSERVICE']._serialized_start = 538
    _globals['_DYNAMICSUBCLASSSERVICE']._serialized_end = 771


import os
import sys

import importlib

import grpc

from model_runner.grpc.generated import train_infer_pb2_grpc
from model_runner.grpc.generated.train_infer_pb2 import InferRequest, InferResponse
from model_runner.grpc.generated.commons_pb2 import Argument, Variant
from model_runner.utils.checkers import ensure_function
from model_runner.utils.datatype_transformer import decode_data, encode_data, detect_data_type
from google.protobuf import empty_pb2


class InferStream:
    def __init__(self):
        self.value = None

    def __iter__(self):
        return self

    def __next__(self):
        if self.value is None:
            raise StopIteration()

        return self.value


class TrainInferStreamServicer(train_infer_pb2_grpc.TrainInferStreamServiceServicer):
    def __init__(self,
                 code_directory: str,
                 resource_directory: str,
                 has_gpu: bool = False,
                 main_file="main.py",
                 ):

        self.main_file = main_file
        self.code_directory = code_directory
        self.resource_directory = resource_directory
        self.has_gpu = has_gpu

        self.module = None
        self.infer_function = None
        self.train_function = None

        self.infer_stream = None
        self.infer_generator = None
        self.stepup = False
        super()

    def Setup(self, request, context):
        if not self.stepup:
            self.module = self.import_code()
            self.infer_function = ensure_function(self.module, "infer")
            self.train_function = ensure_function(self.module, "train")
            self.infer_stream = InferStream()

            self._setup_infer_generator()
            self.stepup = True

        return empty_pb2.Empty()

    def Reinitialize(self, request, context):
        self._setup_infer_generator()
        return empty_pb2.Empty()

    def Infer(self, infer_request: InferRequest, context):
        if self.stepup is None:
            context.abort(code=grpc.StatusCode.FAILED_PRECONDITION, details="Call Setup first")

        try:

            self.infer_stream.value = decode_data(infer_request.argument.value, infer_request.argument.type)

            print(f"InferRequest : {self.infer_stream.value}")

            prediction = next(self.infer_generator)

            print(f"InferResponse : {prediction}")

            # Require prediction type in Setup ??
            type_of_prediction = detect_data_type(prediction)
            infer_response = InferResponse(prediction=Variant(type=type_of_prediction, value=encode_data(type_of_prediction, prediction)))

            return infer_response

        except Exception as e:
            print(str(e))
            self._setup_infer_generator()
            context.abort(code=grpc.StatusCode.INTERNAL, details=e)

    def _setup_infer_generator(self):
        self.infer_generator = self.infer_function(self.infer_stream)
        next(self.infer_generator)

    def import_code(self):
        """
        Import the code from the main_file in the code_directory and return the module object.

        :return: The module object containing the code from the main_file
        """

        sys.path.insert(0, self.code_directory)

        file_path = os.path.join(self.code_directory, self.main_file)
        module_name = 'submission code'
        # Create a module specification from the file path
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        if spec is None:
            raise ImportError(f"Cannot create a spec for module {module_name} from {file_path}", file_path)

        # Create the module from the spec
        module = importlib.util.module_from_spec(spec)

        # Execute the module to load it
        spec.loader.exec_module(module)

        return module



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 29, 0, '', 'commons.proto')
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rcommons.proto\x12\x07commons"<\n\x07Variant\x12"\n\x04type\x18\x02 \x01(\x0e2\x14.commons.VariantType\x12\r\n\x05value\x18\x03 \x01(\x0c"<\n\x08Argument\x12\x10\n\x08position\x18\x01 \x01(\r\x12\x1e\n\x04data\x18\x02 \x01(\x0b2\x10.commons.Variant"=\n\nKwArgument\x12\x0f\n\x07keyword\x18\x01 \x01(\t\x12\x1e\n\x04data\x18\x02 \x01(\x0b2\x10.commons.Variant"\'\n\x06Status\x12\x0c\n\x04code\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t*Z\n\x0bVariantType\x12\x08\n\x04NONE\x10\x00\x12\n\n\x06DOUBLE\x10\x01\x12\x07\n\x03INT\x10\x02\x12\n\n\x06STRING\x10\x03\x12\x0b\n\x07PARQUET\x10\x04\x12\t\n\x05ARROW\x10\x05\x12\x08\n\x04JSON\x10\x06b\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'commons_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_VARIANTTYPE']._serialized_start = 254
    _globals['_VARIANTTYPE']._serialized_end = 344
    _globals['_VARIANT']._serialized_start = 26
    _globals['_VARIANT']._serialized_end = 86
    _globals['_ARGUMENT']._serialized_start = 88
    _globals['_ARGUMENT']._serialized_end = 148
    _globals['_KWARGUMENT']._serialized_start = 150
    _globals['_KWARGUMENT']._serialized_end = 211
    _globals['_STATUS']._serialized_start = 213
    _globals['_STATUS']._serialized_end = 252


from concurrent import futures
import grpc
from model_runner.grpc.generated import dynamic_subclass_pb2_grpc, train_infer_pb2_grpc
from model_runner.servicers.train_infer_servicer import TrainInferStreamServicer
from model_runner.servicers.dynamic_subclass_servicer import DynamicSubclassServicer
import click

import logging

logger = logging.getLogger('model_runner')
logging.basicConfig(level=logging.INFO, format="%(levelname)-8s - %(message)s")

@click.command()
@click.option('--address', default='[::]:50051', envvar='GRPC_ADDRESS', help='IP + Port of server GRPC.')
@click.option('--code-directory', envvar='CODE_DIRECTORY', default='/workspace/submission/code/')
@click.option('--resource-directory', envvar='RESOURCE_DIRECTORY', default='/workspace/resources')
@click.option('--has-gpu', envvar='HAS_GPU', type=bool, default=False, help='Information if GPU is available')
@click.option('--main-file', envvar='MAIN_FILE', default='main.py', help="main file's name of model")
@click.option('--loglevel', default='INFO', envvar='LOG_LEVEL', help='Logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)')
def serve(address, code_directory, resource_directory, has_gpu, main_file, loglevel):
    """Program giving access remotely to model via RPC"""
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    train_infer_pb2_grpc.add_TrainInferStreamServiceServicer_to_server(
        TrainInferStreamServicer(
            code_directory=code_directory,
            resource_directory=resource_directory,
            has_gpu=has_gpu,
            main_file=main_file),
        server)
    dynamic_subclass_pb2_grpc.add_DynamicSubclassServiceServicer_to_server(
        DynamicSubclassServicer(
            code_directory=code_directory
        ),
        server
    )

    logger.setLevel(logging.getLevelName(loglevel.upper()))

    server.add_insecure_port(address)
    logger.info(f'ModelRunner started and ready to serve')
    server.start()
    server.wait_for_termination()



from enum import Enum

import logging

import grpc
import sys

from model_runner.grpc.generated import dynamic_subclass_pb2_grpc
from model_runner.grpc.generated.commons_pb2 import Variant, Status
from model_runner.grpc.generated.dynamic_subclass_pb2 import SetupResponse, SetupRequest, CallRequest, CallResponse, RestResponse
from model_runner.utils import class_resolver
from model_runner.utils.datatype_transformer import decode_data, detect_data_type, encode_data
from google.protobuf import empty_pb2

logger = logging.getLogger(f'model_runner.{__name__}')


class DynamicSubclassStatus(Enum):
    SUCCESS = 'SUCCESS'
    INVALID_ARGUMENT = 'INVALID_ARGUMENT'
    FAILED_PRECONDITION = 'FAILED_PRECONDITION'
    BAD_IMPLEMENTATION = 'BAD_IMPLEMENTATION'
    SETUP_FAILED = 'SETUP_FAILED'
    MODEL_FAILED = 'MODEL_FAILED'
    UNKNOWN = 'UNKNOWN'


class DynamicSubclassServicer(dynamic_subclass_pb2_grpc.DynamicSubclassServiceServicer):
    def __init__(self, code_directory):
        self.code_directory = code_directory
        self.instance = None
        super().__init__()

    def Setup(self, request: SetupRequest, context) -> SetupResponse:
        logger.info('Setup of the model requested')
        if self.instance is not None:
            logger.debug('[Coordinator] Setup has already been called and an instance exists, setup is ignored')
            return SetupResponse(status=Status(code=DynamicSubclassStatus.SUCCESS.name, message='Instance already exists'))

        try:
            class_name = request.className.strip()
            if not class_name:
                logger.error('[Coordinator] Invalid argument, class_name cannot be empty')
                return SetupResponse(status=Status(code=DynamicSubclassStatus.INVALID_ARGUMENT.name, message='class_name cannot be empty'))

            args, kwargs = self.prepare_arguments(request.instanceArguments, request.instanceKwArguments)
            self.instance = class_resolver.load_instance(self.code_directory, class_name, *args, **kwargs)

            logger.info(f'Successfully created instance of class: {self.instance.__class__.__name__} with arguments: {args} and keyword arguments: {kwargs}')
            logger.info('Setup successfully complete')

            return SetupResponse(status=Status(code=DynamicSubclassStatus.SUCCESS.name, message='Instance created successfully'))
        except ImportError as e:
            logger.error('BAD_IMPLEMENTATION: Import error occurred', exc_info=True)
            return SetupResponse(status=Status(code=DynamicSubclassStatus.BAD_IMPLEMENTATION.name, message=e.msg))
        except Exception as e:
            logger.error('SETUP_FAILED: An exception occurred during setup', exc_info=True)
            return SetupResponse(status=Status(code=DynamicSubclassStatus.SETUP_FAILED.name, message=str(e)))

    def Call(self, request: CallRequest, context) -> CallResponse | None:
        if self.instance is None:
            logger.error('[Coordinator] FAILED_PRECONDITION - Setup has not been called yet')
            return CallResponse(
                status=Status(code=DynamicSubclassStatus.FAILED_PRECONDITION.name, message='Setup has not been called yet')
            )

        method_name = request.methodName
        if method_name == '':
            logger.error('[Coordinator] INVALID_ARGUMENT - methodName cannot be empty')
            return CallResponse(
                status=Status(code=DynamicSubclassStatus.INVALID_ARGUMENT.name, message='methodName cannot be empty')
            )

        method = None
        try:
            method = getattr(self.instance, method_name)
        except AttributeError as e:
            logger.error(f'BAD_IMPLEMENTATION: Method "{method_name}" not found in class "{self.instance.__class__.__name__}"')
            return CallResponse(
                status=Status(
                    code=DynamicSubclassStatus.BAD_IMPLEMENTATION.name,
                    message=f'Method "{method_name}" not found in class "{self.instance.__class__.__name__}"'
                )
            )

        try:
            args, kwargs = self.prepare_arguments(request.methodArguments, request.methodKwArguments)

            logger.debug('Call to method "%s" with positional arguments: %s, keyword arguments: %s', method_name, args, kwargs)
            method_result = method(*args, **kwargs)
            logger.debug('Model response: %s', method_result)
            if method_result is None:
                return CallResponse(status=Status(code=DynamicSubclassStatus.SUCCESS.name, message=''))

            type_of_result = detect_data_type(method_result)
            logger.debug("Detected type of result: %s", type_of_result)

            encoded_result: bytes = encode_data(type_of_result, method_result)
            return CallResponse(
                status=Status(code=DynamicSubclassStatus.SUCCESS.name, message=''),
                methodResponse=Variant(type=type_of_result, value=encoded_result)
            )

        except Exception as e:
            logger.error(f'INTERNAL: The model raised an exception', exc_info=True)
            return CallResponse(
                status=Status(code=DynamicSubclassStatus.MODEL_FAILED.name, message=f'The model raised an exception: {str(e)}')
            )

    def Rest(self, request: empty_pb2.Empty, context) -> RestResponse:
        logger.info('[Coordinator] Resetting the current instance')
        if self.instance is None:
            logger.warning('[Coordinator] No existing instance to reset')
            return RestResponse(
                status=Status(
                    code=DynamicSubclassStatus.SUCCESS.name,
                    message='No instance exists to reset'
                )
            )

        try:
            self.instance = None
            logger.info('[Coordinator] Instance successfully reset')
            return RestResponse(
                status=Status(
                    code=DynamicSubclassStatus.SUCCESS.name,
                    message='Instance successfully reset'
                )
            )
        except Exception as e:
            logger.error('[Coordinator] INTERNAL: Failed to reset instance', exc_info=True)
            return RestResponse(
                status=Status(
                    code=DynamicSubclassStatus.UNKNOWN.name,
                    message=f'Failed to reset instance: {str(e)}'
                )
            )

    @staticmethod
    def prepare_arguments(args, kwargs):
        args = [
            decode_data(arg.data.value, arg.data.type) for arg in sorted(args, key=lambda arg: arg.position)
        ]

        kwargs = {
            kwarg.keyword: decode_data(kwarg.data.value, kwarg.data.type) for kwarg in kwargs
        }

        return args, kwargs



import commons_pb2 as _commons_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class SetupRequest(_message.Message):
    __slots__ = ('className', 'instanceArguments', 'instanceKwArguments')
    CLASSNAME_FIELD_NUMBER: _ClassVar[int]
    INSTANCEARGUMENTS_FIELD_NUMBER: _ClassVar[int]
    INSTANCEKWARGUMENTS_FIELD_NUMBER: _ClassVar[int]
    className: str
    instanceArguments: _containers.RepeatedCompositeFieldContainer[_commons_pb2.Argument]
    instanceKwArguments: _containers.RepeatedCompositeFieldContainer[_commons_pb2.KwArgument]

    def __init__(self, className: _Optional[str]=..., instanceArguments: _Optional[_Iterable[_Union[_commons_pb2.Argument, _Mapping]]]=..., instanceKwArguments: _Optional[_Iterable[_Union[_commons_pb2.KwArgument, _Mapping]]]=...) -> None:
        ...

class SetupResponse(_message.Message):
    __slots__ = ('status',)
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _commons_pb2.Status

    def __init__(self, status: _Optional[_Union[_commons_pb2.Status, _Mapping]]=...) -> None:
        ...

class CallRequest(_message.Message):
    __slots__ = ('methodName', 'methodArguments', 'methodKwArguments')
    METHODNAME_FIELD_NUMBER: _ClassVar[int]
    METHODARGUMENTS_FIELD_NUMBER: _ClassVar[int]
    METHODKWARGUMENTS_FIELD_NUMBER: _ClassVar[int]
    methodName: str
    methodArguments: _containers.RepeatedCompositeFieldContainer[_commons_pb2.Argument]
    methodKwArguments: _containers.RepeatedCompositeFieldContainer[_commons_pb2.KwArgument]

    def __init__(self, methodName: _Optional[str]=..., methodArguments: _Optional[_Iterable[_Union[_commons_pb2.Argument, _Mapping]]]=..., methodKwArguments: _Optional[_Iterable[_Union[_commons_pb2.KwArgument, _Mapping]]]=...) -> None:
        ...

class CallResponse(_message.Message):
    __slots__ = ('status', 'methodResponse')
    STATUS_FIELD_NUMBER: _ClassVar[int]
    METHODRESPONSE_FIELD_NUMBER: _ClassVar[int]
    status: _commons_pb2.Status
    methodResponse: _commons_pb2.Variant

    def __init__(self, status: _Optional[_Union[_commons_pb2.Status, _Mapping]]=..., methodResponse: _Optional[_Union[_commons_pb2.Variant, _Mapping]]=...) -> None:
        ...

class RestResponse(_message.Message):
    __slots__ = ('status',)
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _commons_pb2.Status

    def __init__(self, status: _Optional[_Union[_commons_pb2.Status, _Mapping]]=...) -> None:
        ...


"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import train_infer_pb2 as train__infer__pb2
GRPC_GENERATED_VERSION = '1.69.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False
try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True
if _version_not_supported:
    raise RuntimeError(f'The grpc package installed is at version {GRPC_VERSION},' + f' but the generated code in train_infer_pb2_grpc.py depends on' + f' grpcio>={GRPC_GENERATED_VERSION}.' + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}' + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.')

class TrainInferStreamServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Setup = channel.unary_unary('/train_infer.TrainInferStreamService/Setup', request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, _registered_method=True)
        self.Infer = channel.unary_unary('/train_infer.TrainInferStreamService/Infer', request_serializer=train__infer__pb2.InferRequest.SerializeToString, response_deserializer=train__infer__pb2.InferResponse.FromString, _registered_method=True)
        self.Reinitialize = channel.unary_unary('/train_infer.TrainInferStreamService/Reinitialize', request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, _registered_method=True)

class TrainInferStreamServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Setup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Infer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Reinitialize(self, request, context):
        """rpc Train(TrainRequest) returns (TrainResponse);
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

def add_TrainInferStreamServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {'Setup': grpc.unary_unary_rpc_method_handler(servicer.Setup, request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString), 'Infer': grpc.unary_unary_rpc_method_handler(servicer.Infer, request_deserializer=train__infer__pb2.InferRequest.FromString, response_serializer=train__infer__pb2.InferResponse.SerializeToString), 'Reinitialize': grpc.unary_unary_rpc_method_handler(servicer.Reinitialize, request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString)}
    generic_handler = grpc.method_handlers_generic_handler('train_infer.TrainInferStreamService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('train_infer.TrainInferStreamService', rpc_method_handlers)

class TrainInferStreamService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Setup(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/train_infer.TrainInferStreamService/Setup', google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, google_dot_protobuf_dot_empty__pb2.Empty.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def Infer(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/train_infer.TrainInferStreamService/Infer', train__infer__pb2.InferRequest.SerializeToString, train__infer__pb2.InferResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def Reinitialize(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/train_infer.TrainInferStreamService/Reinitialize', google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, google_dot_protobuf_dot_empty__pb2.Empty.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

class TrainInferServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Setup = channel.unary_unary('/train_infer.TrainInferService/Setup', request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, _registered_method=True)
        self.Infer = channel.unary_unary('/train_infer.TrainInferService/Infer', request_serializer=train__infer__pb2.InferRequest.SerializeToString, response_deserializer=train__infer__pb2.InferResponse.FromString, _registered_method=True)
        self.Reinitialize = channel.unary_unary('/train_infer.TrainInferService/Reinitialize', request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, _registered_method=True)

class TrainInferServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Setup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Infer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Reinitialize(self, request, context):
        """rpc Train(TrainRequest) returns (TrainResponse);
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

def add_TrainInferServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {'Setup': grpc.unary_unary_rpc_method_handler(servicer.Setup, request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString), 'Infer': grpc.unary_unary_rpc_method_handler(servicer.Infer, request_deserializer=train__infer__pb2.InferRequest.FromString, response_serializer=train__infer__pb2.InferResponse.SerializeToString), 'Reinitialize': grpc.unary_unary_rpc_method_handler(servicer.Reinitialize, request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString, response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString)}
    generic_handler = grpc.method_handlers_generic_handler('train_infer.TrainInferService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('train_infer.TrainInferService', rpc_method_handlers)

class TrainInferService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Setup(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/train_infer.TrainInferService/Setup', google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, google_dot_protobuf_dot_empty__pb2.Empty.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def Infer(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/train_infer.TrainInferService/Infer', train__infer__pb2.InferRequest.SerializeToString, train__infer__pb2.InferResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def Reinitialize(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/train_infer.TrainInferService/Reinitialize', google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString, google_dot_protobuf_dot_empty__pb2.Empty.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)


import logging

import importlib
import inspect
import pkgutil

logger = logging.getLogger(f'model_runner.{__name__}')

def load_instance(code_path: str, base_class_name: str, *args, **kwargs):
    """
    Dynamically searches for and instantiates a class by its name, assuming it inherits from a base class.
    :param code_path: The path to the directory containing the code.
    :param base_class_name: The full name of the base class.
    :param args: Positional arguments to pass to the class constructor.
    :param kwargs: Keyword arguments to pass to the class constructor.
    :return: An instance of the class if successfully found and instantiated.
    :raises ImportError: If the class cannot be found or does not inherit from the base class.
    """
    import sys
    sys.path.append(code_path)

    logger.info(f"Loading class '{base_class_name}' from '{code_path}'")
    # todo: maybe is not required to walk packages and only import the root module ?
    base_class = resolve_class(base_class_name)
    for importer, module_name, is_package in pkgutil.walk_packages([code_path]):
        try:
            module = importlib.import_module(module_name)
        except Exception as e:
            logger.error(f"Error importing module '{module_name}'", exc_info=True)
            continue

        for _, obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(obj, (base_class)) and obj is not base_class:
                logger.info(f"Found class '{obj.__name__}' that inherits from '{base_class.__name__}'.")
                return obj(*args, **kwargs)
            else:
                logger.debug(f"Class '{obj.__name__}' does not inherit from '{base_class.__name__}'.")

    raise ImportError(f"No Inherited class found from '{base_class}'.")


def resolve_class(class_full_name: str):
    module_name, _, class_name = class_full_name.rpartition('.')
    if not module_name:
        raise ValueError(f"Invalid class name '{class_name}'. Use 'module.ClassName' format.")
    module = importlib.import_module(module_name)
    if not hasattr(module, class_name):
        raise ValueError(f"Class '{class_name}' not found in module '{module_name}'.")
    class_obj = getattr(module, class_name)
    if not inspect.isclass(class_obj):
        raise ValueError(f"Object '{class_name}' in module '{module_name}' is not a class.")

    return class_obj
